/*
    This file is part of the KDE project
    SPDX-FileCopyrightText: 2010 Sebastian Trueg <trueg@kde.org>

    SPDX-License-Identifier: LGPL-2.0-only OR LGPL-3.0-only OR LicenseRef-KDE-Accepted-LGPL
*/

#include "kabstractfileitemactionplugin.h"

KAbstractFileItemActionPlugin::KAbstractFileItemActionPlugin(QObject *parent)
    : QObject(parent)
{
}

KAbstractFileItemActionPlugin::~KAbstractFileItemActionPlugin()
{
}

#include "moc_kabstractfileitemactionplugin.cpp"
